#ifndef DUNE_MICROSTRUCTURE_PRESTRAIN_MATERIAL_GEOMETRY_HH
#define DUNE_MICROSTRUCTURE_PRESTRAIN_MATERIAL_GEOMETRY_HH


#include <dune/grid/uggrid.hh>
#include <dune/grid/yaspgrid.hh>
#include <dune/microstructure/matrix_operations.hh>

using namespace Dune;
using namespace MatrixOperations;
using std::pow;
using std::abs;
using std::sqrt;
using std::sin;
using std::cos;





template <int dim> 
class IsotropicMaterialImp 
{

public:
	static const int dimWorld = dim;
    using CellGridType = YaspGrid< dim, EquidistantOffsetCoordinates< double, dim>>;
    using Domain = typename CellGridType::LeafGridView::template Codim<0>::Geometry::GlobalCoordinate;
    using MatrixRT = FieldMatrix< double, dimWorld, dimWorld>;
    using Func2Tensor = std::function< MatrixRT(const Domain&) >;
	using FuncScalar = std::function< double(const Domain&) >;


    FuncScalar getMu(ParameterTree parameters){

    	std::string imp = parameters.get<std::string>("material_prestrain_imp", "analytical_Example");
    	//std::array<std::string,5> imps({"homogen_poisson" "bilayer_poisson" "chess_poisson" "3Dchess_poisson" "vertical_bilayer_poisson"});
        
    	double phi = M_PI*parameters.get<double>("material_angle", 0.0)/180; //TODO 
    	double theta = parameters.get<double>("theta",1.0/4.0);                        //TODO alle parameter hier laden?
        
        double width = parameters.get<double>("width", 1.0);
        double height = parameters.get<double>("height", 1.0);


    	if (imp == "homogeneous"){    
		    double mu1     = parameters.get<double>("mu1", 10);

		    auto muTerm = [mu1] (const Domain& z) {return mu1;};
		    
			return muTerm;
		}
		else if (imp == "isotropic_bilayer")
    	{
            double mu1     = parameters.get<double>("mu1",10.0);
            double beta = parameters.get<double>("beta",2.0); 
		    double mu2 = beta*mu1;
    
 		   auto muTerm = [mu1, mu2, theta] (const Domain& z) {
                    if (z[2]>0)                                                   
                        return mu1;
                    else
                        return mu2;
                    };
		    
           return muTerm;
		}
        else if (imp == "analytical_Example"){	
		    double mu1     = parameters.get<double>("mu1",10.0);
            double beta = parameters.get<double>("beta",2.0); 
		    double mu2 = beta*mu1;
		    
		   auto muTerm = [mu1, mu2, theta] (const Domain& z) {
//                     std::cout << "Analytical-MU is used" << std::endl;
                    if (abs(z[0]) >= (theta/2.0))                                                   
                        return mu1;
                    else
                        return mu2;
                    };
		    
           return muTerm;
		}
        else if (imp == "parametrized_Laminate"){	
		    double mu1     = parameters.get<double>("mu1",1.0);
            double beta = parameters.get<double>("beta",2.0); 
		    double mu2 = beta*mu1;
		    
		   auto muTerm = [mu1, mu2, theta] (const Domain& z) {
//                     std::cout << "Analytical-MU is used" << std::endl;
//                     if (abs(z[0]) >= (theta/2.0))      
                    if (abs(z[0]) > (theta/2.0))  
                        return mu1;
                    else
                        return mu2;
                    };
		    
           return muTerm;
		}
		else if (imp == "circle_fiber"){
		    double mu1     = parameters.get<double>("mu1",10.0);
            double beta = parameters.get<double>("beta",2.0); 
		    double mu2 = beta*mu1;

	 		auto muTerm = [mu1,mu2,width] (const Domain& z) { 
		      	if (z[0] < 0 && sqrt(pow(z[1],2) + pow(z[2],2) ) < width/4.0  ||  0 < z[0] && sqrt(pow(z[1],2) + pow(z[2],2) ) > width/4.0)
		        	return mu1;
		      	else
		        	return mu2;
		    };
            
            return muTerm;
   		}
        else if (imp == "square_fiber"){
		    double mu1     = parameters.get<double>("mu1",10.0);
            double beta = parameters.get<double>("beta",2.0); 
		    double mu2 = beta*mu1;

	 		auto muTerm = [mu1,mu2,width] (const Domain& z) { 
                if (z[0] < 0 && std::max(abs(z[1]), abs(z[2]))  < width/4.0  ||  0 < z[0] && std::max(abs(z[1]), abs(z[2]))  > width/4.0) 
		        	return mu1;
		      	else
		        	return mu2;
		    };
            
            return muTerm;
   		}
        else if (imp == "matrix_material_circles")   // Matrix material with prestrained Fiber inclusions (circular cross-section)
        {
            double mu1     = parameters.get<double>("mu1",10.0);
            double beta    = parameters.get<double>("beta",2.0); 
		    double mu2     = beta*mu1;
            
            int nF    = parameters.get<int>("nF", 2);    //number of Fibers in each Layer
            double rF = parameters.get<double>("rF", 0.5*(width/(2.0*nF)) );   //default: half of the max-fiber-radius mrF = (width/(2.0*nF)) 
            
            
            
            assert( (2*rF)*nF <= width && (height/4)+rF <= height); //check that fibers are not bigger than Domain
            
            auto muTerm = [mu1,mu2,theta,nF,rF,height,width] (const Domain& x)                
            {              
                //TODO if Domain == 1... if Domain == 2 ...
//                 double domainShift = 1.0/2.0;
                // TEST
                double domainShift = 0.0;
                // shift x to domain [0,1]^3
                FieldVector<double,3> s = {x[0]+domainShift, x[1]+domainShift, x[2]+domainShift};
//                 std::cout << "matrixMaterial-MU is used" << std::endl;
                
                double output;
                
                // determine if point is in upper/lower Layer
                if ( 0  <= s[2] && s[2] <= 1.0/2.0) // upper Layer
                {
                    for(size_t i=0; i<nF ;i++) // running through all the Fibers.. 
                    {
                        if(-1.0/2.0 + (1.0/nF)*i<= s[0] && s[0] <= -1.0/2.0 +  (1.0/double(nF))*(i+1))     // have to evenly space fibers... 
                        {
//                             std::cout << " i : " << i << std::endl;
//                             printvector(std::cout, s, "s" , "--");
                            // compute Fiber center
                            FieldVector<double,2> Fcenter = { (1.0/(2.0*nF))+((1.0/double(nF))*i)-(1.0/2.0) , 1.0/4.0};
//                             printvector(std::cout, Fcenter, "Fcenter" , "--");
//                             
                            if(sqrt(pow(s[0]-Fcenter[0],2)+pow(s[2]-Fcenter[1],2)) <= rF )
                                output = mu2;
//                                 return mu2;                                                         //richtig so? Fiber hat tendenziell größeres mu? 
                            else
                                output = mu1;
//                                 return mu1;
                        }
                        else {}
                    }
                }
                else  // lower Layer
                {
                    for(size_t i=0; i<nF ;i++) // running through all the Fibers.. 
                    {
                        if(-1.0/2.0 + (1.0/double(nF))*i<= s[0] && s[0] <= -1.0/2.0 +  (1.0/double(nF))*(i+1))  
                        {
//                             std::cout << " i : " << i << std::endl;
//                             printvector(std::cout, s, "s" , "--");
                            // compute Fiber center
                            FieldVector<double,2> Fcenter = { (1.0/(2.0*nF))+((1.0/double(nF))*i)-(1.0/2.0) , -1.0/4.0 };
//                             printvector(std::cout, Fcenter, "Fcenter" , "--");
                            
                            if(sqrt(pow(s[0]-Fcenter[0],2)+pow(s[2]-Fcenter[1],2)) <= rF )
                                output = mu2;
//                                 return mu2;                                                         //richtig so? Fiber hat tendenziell größeres mu? 
                            else
                                output = mu1;
//                                 return mu1;
                        }
                        else{}
                    }

                }
                return output;
            };
            return muTerm;
        }
        else if (imp == "matrix_material_squares")    // Matrix material with prestrained Fiber inclusions (square-shaped cross-section)
        {
            double mu1     = parameters.get<double>("mu1",10.0);
            double beta    = parameters.get<double>("beta",2.0); 
		    double mu2     = beta*mu1;
            
            int nF    = parameters.get<int>("nF", 2);    //number of Fibers in each Layer
            double rF = parameters.get<double>("rF", 0.5*(width/(2.0*nF)) );   //default: half of the max-fiber-radius mrF = (width/(2.0*nF)) 
            
            
            
            assert( (2*rF)*nF <= width && (height/4)+rF <= height); //check that fibers are not bigger than Domain
            
            auto muTerm = [mu1,mu2,theta,nF,rF,height,width] (const Domain& x)                
            {              
                //TODO if Domain == 1... if Domain == 2 ...
//                 double domainShift = 1.0/2.0;
                // TEST
                double domainShift = 0.0;
                // shift x to domain [0,1]^3
                FieldVector<double,3> s = {x[0]+domainShift, x[1]+domainShift, x[2]+domainShift};
//                 std::cout << "matrixMaterial-MU is used" << std::endl;
                double output;
                
                // determine if point is in upper/lower Layer
                if ( 0  <= s[2] && s[2] <= 1.0/2.0) // upper Layer
                {
                    for(size_t i=0; i<nF ;i++) // running through all the Fibers.. 
                    {
                        if(-1.0/2.0 + (1.0/nF)*i<= s[0] && s[0] <= -1.0/2.0 +  (1.0/double(nF))*(i+1))     // have to evenly space fibers... 
                        {
//                             std::cout << " i : " << i << std::endl;
//                             printvector(std::cout, s, "s" , "--");
                            // compute Fiber center
                            FieldVector<double,2> Fcenter = { (1.0/(2.0*nF))+((1.0/double(nF))*i)-(1.0/2.0) , 1.0/4.0};
//                             printvector(std::cout, Fcenter, "Fcenter" , "--");
//                             
                            if(std::max( abs(s[0]-Fcenter[0]), abs(s[2]-Fcenter[1])  ) <= rF )
                                output = mu2;
//                                 return mu2;                                                         //richtig so? Fiber hat tendenziell größeres mu? 
                            else
                                output = mu1;
//                                 return mu1;
                        }
                    }
                }
                else  // lower Layer
                {
                    for(size_t i=0; i<nF ;i++) // running through all the Fibers.. 
                    {
                        if(-1.0/2.0 + (1.0/double(nF))*i<= s[0] && s[0] <= -1.0/2.0 +  (1.0/double(nF))*(i+1))  
                        {
//                             std::cout << " i : " << i << std::endl;
//                             printvector(std::cout, s, "s" , "--");
                            // compute Fiber center
                            FieldVector<double,2> Fcenter = { (1.0/(2.0*nF))+((1.0/double(nF))*i)-(1.0/2.0) , -1.0/4.0 };
//                             printvector(std::cout, Fcenter, "Fcenter" , "--");
                            
                            if(std::max( abs(s[0]-Fcenter[0]), abs(s[2]-Fcenter[1])  ) <= rF )
                                output = mu2;
//                                 return mu2;                                                         //richtig so? Fiber hat tendenziell größeres mu? 
                            else
                                output = mu1;
//                                 return mu1;
                        }
                    }
                }
                return output;
            };
            return muTerm;
        }
        else if (imp == "bilayer"){	
		    double mu1     = parameters.get<double>("mu1",10.0);
            double beta = parameters.get<double>("beta",2.0); 
		    double mu2 = beta*mu1;
		    
		    auto muTerm = [mu1, mu2, phi] (const Domain& z) {
		      if (isInRotatedPlane(phi, z[dim-2], z[dim-1]))
		        return mu1;
		      else
		        return mu2; 
            };
		    
			return muTerm;
		}
		


		else if (imp == "helix_poisson"){                                         // interessant!
		    double mu1     = parameters.get<double>("mu1",10.0);
            double beta    = parameters.get<double>("beta",2.0); 
		    double mu2     = beta*mu1;

		    double r  = parameters.get<double>("radius");
			double r2 = parameters.get<double>("radius_helix");

		    auto muTerm = [mu1, mu2, r, r2] (const Domain& z) {
		    	std::array<double,2> h = {r*cos(2*M_PI*z[0]), r*sin(2*M_PI*z[0])};
		      	if ( sqrt(pow(z[1]-h[0],2)+pow(z[2]-h[1],2)) < r2 )                   // distance to the helix?
		        	return mu1;
		      	else
		        	return mu2; 
		    };
		    
			return muTerm;
		}

		else if (imp == "chess_poisson"){
		    double mu1     = parameters.get<double>("mu1",10.0);
            double beta    = parameters.get<double>("beta",2.0); 
		    double mu2     = beta*mu1;

		    auto muTerm = [mu1, mu2, phi] (const Domain& z) {
		      	if ( (isInRotatedPlane(phi, 	   z[dim-2], z[dim-1]) and isInRotatedPlane(phi + M_PI/2  , z[dim-2], z[dim-1])) or 
		      		 (isInRotatedPlane(phi + M_PI, z[dim-2], z[dim-1]) and isInRotatedPlane(phi + 3*M_PI/2, z[dim-2], z[dim-1])) )
		        	return mu1;
		      	else
		        	return mu2; 
		    };
		    
			return muTerm;
		}

// 		else if (imp == "3Dchess_poisson"){
// 		    double E1                = parameters.get<double>("E1", 17e6); //Faser
// 		    double nu1               = parameters.get<double>("nu1", 0.3);
// 		    double mu1 = lameMu(E1, nu1);
// 
// 		    double E2                = parameters.get<double>("E2", 17e6); //Polymer, weicher, Querkontraktion groß
// 		    double nu2               = parameters.get<double>("nu2", 0.5); 
// 		    double mu2 = lameMu(E2, nu2);
// 
// 		    auto muTerm = [mu1, mu2, phi] (const Domain& z) {
// 		    	if (z[0]<0.5)
// 		      		if ( (isInRotatedPlane(phi, 	   z[1], z[2]) and isInRotatedPlane(phi + M_PI/2  , z[1], z[2])) or 
// 		      			 (isInRotatedPlane(phi + M_PI, z[1], z[2]) and isInRotatedPlane(phi + 3*M_PI/2, z[1], z[2])) )
// 		        		return mu1;
// 		      		else
// 		        		return mu2; 
// 		        else
// 		        	if ( (isInRotatedPlane(phi,  	   z[1], z[2]) and isInRotatedPlane(phi + M_PI/2  , z[1], z[2])) or 
// 		      			 (isInRotatedPlane(phi + M_PI, z[1], z[2]) and isInRotatedPlane(phi + 3*M_PI/2, z[1], z[2])) )
// 		        		return mu2;
// 		      		else
// 		        		return mu1; 
// 		    };
// 		    
// 			return muTerm;
// 		}
// 		else if (imp == "vertical_bilayer_poisson"){ 
//     		double E1                = parameters.get<double>("E1", 17e6); //material1
// 		    double nu1               = parameters.get<double>("nu1", 0.3);
// 		    double mu1 = lameMu(E1, nu1);
// 		    double E2                = parameters.get<double>("E2", 17e6); //material2
// 		    double nu2               = parameters.get<double>("nu2", 0.5); 
// 		    double mu2 = lameMu(E2, nu2);
// 
// 		    auto muTerm = [mu1, mu2, phi] (const Domain& z) {
// 		      if ( isInRotatedPlane(phi+M_PI/2.0, z[dim-2], z[dim-1]) )//x3
// 		        return mu1;
// 		      else
// 		        return mu2; };
// 		    
// 			return muTerm;
// 		}
// 
// 		else if (imp == "microstructured_bilayer_poisson"){ 
//     		double E1                = parameters.get<double>("E1", 17e6); //material1
// 		    double nu1               = parameters.get<double>("nu1", 0.3);
// 		    double mu1 = lameMu(E1, nu1);
// 		    double E2                = parameters.get<double>("E2", 17e6); //material2
// 		    double nu2               = parameters.get<double>("nu2", 0.5); 
// 		    double mu2 = lameMu(E2, nu2);
// 
// 		    auto muTerm = [mu1, mu2, phi] (const Domain& z) {
// 		      if ( isInRotatedPlane(phi, z[0]-0.5, z[1]) )                 //TODO understand this
// 		        return mu1;
// 		      else
// 		        return mu2; };
// 		    
// 			return muTerm;
// 		}
// 
// 		else { //(imp == "bilayer_poisson")
//     		double E1                = parameters.get<double>("E1", 17e6); //material1
// 		    double nu1               = parameters.get<double>("nu1", 0.3);
// 		    double mu1 = lameMu(E1, nu1);
// 		    double E2                = parameters.get<double>("E2", 17e6); //material2
// 		    double nu2               = parameters.get<double>("nu2", 0.5); 
// 		    double mu2 = lameMu(E2, nu2);
// 
// 		    auto muTerm = [mu1, mu2, phi] (const Domain& z) {
// 		      if ( isInRotatedPlane(phi, z[dim-2], z[dim-1]) ) //x2
// 		        return mu1;
// 		      else
// 		        return mu2; };
// 		    
// 			return muTerm;
// 		}
	}

	FuncScalar getLambda(ParameterTree parameters)
    {

		std::string imp = parameters.get<std::string>("material_prestrain_imp", "analytical_Example");
		//std::array<std::string,5> imps({"homogen_poisson" "bilayer_poisson" "chess_poisson" "3Dchess_poisson" "vertical_bilayer_poisson"});
    	//int i_imp =  parameters.get<int>("impnumber");
    	//std::string imp =  imps[i_imp];
    	double phi = M_PI*parameters.get<double>("material_angle", 0.0)/180;
        double theta = parameters.get<double>("theta",1.0/4.0);                        //TODO alle parameter hier laden?
        
        double width = parameters.get<double>("width", 1.0);
        double height = parameters.get<double>("height", 1.0);

		if (imp == "homogenenous"){    
		    double lambda     = parameters.get<double>("lambda",384.615);

		    auto lambdaTerm = [lambda] (const Domain& z) {return lambda;};
		    
			return lambdaTerm;
		}
        else if (imp == "isotropic_bilayer")
    	{
            double lambda1 = parameters.get<double>("lambda1",0.0);
            double beta = parameters.get<double>("beta",2.0); 
            double lambda2 = beta*lambda1;
    
 		   auto lambdaTerm = [lambda1,lambda2, theta] (const Domain& z) {
                    if (z[2]>0)                                                   
                        return lambda1;
                    else
                        return lambda2;
                    };
           return lambdaTerm;
		}

        else if (imp == "analytical_Example"){
            double lambda1 = parameters.get<double>("lambda1",0.0);
            double beta = parameters.get<double>("beta",2.0); 
            double lambda2 = beta*lambda1;
    
            auto lambdaTerm = [lambda1,lambda2, theta] (const Domain& z) {
                
//                  std::cout << "Analytical-LAMBDA is used" << std::endl;  //TEST
                    if (abs(z[0]) >= (theta/2.0))
                        return lambda1;
                    else
                        return lambda2;
                    }; 
           return lambdaTerm;
		}
        else if (imp == "parametrized_Laminate"){
            double lambda1 = parameters.get<double>("lambda1",0.0);
            double beta = parameters.get<double>("beta",2.0); 
            double lambda2 = beta*lambda1;
    
            auto lambdaTerm = [lambda1,lambda2, theta] (const Domain& z) {
                
//                  std::cout << "Analytical-LAMBDA is used" << std::endl;  //TEST
//                     if (abs(z[0]) > (theta/2.0))
                    if (abs(z[0]) > (theta/2.0))
                        return lambda1;
                    else
                        return lambda2;
                    }; 
           return lambdaTerm;
		}
		else if (imp == "circle_fiber"){
            double lambda1 = parameters.get<double>("lambda1",0.0);
            double beta = parameters.get<double>("beta",2.0); 
            double lambda2 = beta*lambda1;

	 		auto lambdaTerm = [lambda1,lambda2,width] (const Domain& z) // Prestrain inducing twist (bisher nur extension)
			{ 
		      	if (z[0] < 0 && sqrt(pow(z[1],2) + pow(z[2],2) ) < width/4.0  ||  0 < z[0] && sqrt(pow(z[1],2) + pow(z[2],2) ) > width/4.0)
		        	return lambda1;
		      	else
		        	return lambda2;
		    };
            
           return lambdaTerm;
   		}
   		else if (imp == "square_fiber"){
            double lambda1 = parameters.get<double>("lambda1",0.0);
            double beta = parameters.get<double>("beta",2.0); 
            double lambda2 = beta*lambda1;

	 		auto lambdaTerm = [lambda1,lambda2,width] (const Domain& z) // Prestrain inducing twist (bisher nur extension)
			{ 
		      	if (z[0] < 0 && std::max(abs(z[1]), abs(z[2]))  < width/4.0  ||  0 < z[0] && std::max(abs(z[1]), abs(z[2]))  > width/4.0) 
		        	return lambda1;
		      	else
		        	return lambda2;
		    };
            
           return lambdaTerm;
   		}
        else if (imp == "matrix_material_circles")   // Matrix material with prestrained Fiber inclusions (circular cross-section)
        {
            double lambda1 = parameters.get<double>("lambda1",0.0);
            double beta = parameters.get<double>("beta",2.0); 
            double lambda2 = beta*lambda1;
            
            int nF    = parameters.get<int>("nF", 2);    //number of Fibers in each Layer
            double rF = parameters.get<double>("rF", 0.5*(width/(2.0*nF)) );   //default: half of the max-fiber-radius mrF = (width/(2.0*nF)) 
            
            

            assert( (2*rF)*nF <= width && (height/4)+rF <= height); //check that fibers are not bigger than Domain
            
            auto lambdaTerm = [lambda1,lambda2, theta,nF,rF,height,width] (const Domain& x)               
            {              
                //TODO if Domain == 1... if Domain == 2 ...
//                 double domainShift = 1.0/2.0;
                // TEST
                double domainShift = 0.0;
                // shift x to domain [0,1]^3
                FieldVector<double,3> s = {x[0]+domainShift, x[1]+domainShift, x[2]+domainShift};
//                 std::cout << "matrixMaterial-MU is used" << std::endl;
                double output;
                
                // determine if point is in upper/lower Layer
                if ( 0  <= s[2] && s[2] <= 1.0/2.0) // upper Layer
                {
                    for(size_t i=0; i<nF ;i++) // running through all the Fibers.. 
                    {
                        if(-1.0/2.0 + (1.0/nF)*i<= s[0] && s[0] <= -1.0/2.0 +  (1.0/double(nF))*(i+1))     // have to evenly space fibers... 
                        {
//                             std::cout << " i : " << i << std::endl;
//                             printvector(std::cout, s, "s" , "--");
                            // compute Fiber center
                            FieldVector<double,2> Fcenter = { (1.0/(2.0*nF))+((1.0/double(nF))*i)-(1.0/2.0) , 1.0/4.0};
//                             printvector(std::cout, Fcenter, "Fcenter" , "--");
//                             
                            if(sqrt(pow(s[0]-Fcenter[0],2)+pow(s[2]-Fcenter[1],2)) <= rF )
                                output = lambda2;
//                                 return lambda2;                                                         //richtig so? Fiber hat tendenziell größeres mu? 
                            else
                                output = lambda1;
//                                 return lambda1;
                        }
                    }
                }
                else  // lower Layer
                {
                    for(size_t i=0; i<nF ;i++) // running through all the Fibers.. 
                    {
                        if(-1.0/2.0 + (1.0/double(nF))*i<= s[0] && s[0] <= -1.0/2.0 +  (1.0/double(nF))*(i+1))  
                        {
//                             std::cout << " i : " << i << std::endl;
//                             printvector(std::cout, s, "s" , "--");
                            // compute Fiber center
                            FieldVector<double,2> Fcenter = { (1.0/(2.0*nF))+((1.0/double(nF))*i)-(1.0/2.0) , -1.0/4.0 };
//                             printvector(std::cout, Fcenter, "Fcenter" , "--");
                            
                            if(sqrt(pow(s[0]-Fcenter[0],2)+pow(s[2]-Fcenter[1],2)) <= rF )
                                output = lambda2;
//                                 return lambda2;                                                         //richtig so? Fiber hat tendenziell größeres mu? 
                            else
                                output = lambda1;
//                                 return lambda1;
                        }
                    }

                }
                return output;
            };
            return lambdaTerm;
        }
        else if (imp == "matrix_material_squares")    // Matrix material with prestrained Fiber inclusions (square-shaped cross-section)
        {
            double lambda1 = parameters.get<double>("lambda1",0.0);
            double beta = parameters.get<double>("beta",2.0); 
            double lambda2 = beta*lambda1;
            
            int nF    = parameters.get<int>("nF", 2);    //number of Fibers in each Layer
            double rF = parameters.get<double>("rF", 0.5*(width/(2.0*nF)) );   //default: half of the max-fiber-radius mrF = (width/(2.0*nF)) 
            
            assert( (2*rF)*nF <= width && (height/4)+rF <= height); //check that fibers are not bigger than Domain
            
            auto lambdaTerm = [lambda1,lambda2, theta,nF,rF,height,width] (const Domain& x)                
            {              
                //TODO if Domain == 1... if Domain == 2 ...
//                 double domainShift = 1.0/2.0;
                // TEST
                double domainShift = 0.0;
                // shift x to domain [0,1]^3
                FieldVector<double,3> s = {x[0]+domainShift, x[1]+domainShift, x[2]+domainShift};
//                 std::cout << "matrixMaterial-MU is used" << std::endl;
                double output;
                
                // determine if point is in upper/lower Layer
                if ( 0  <= s[2] && s[2] <= 1.0/2.0) // upper Layer
                {
                    for(size_t i=0; i<nF ;i++) // running through all the Fibers.. 
                    {
                        if(-1.0/2.0 + (1.0/nF)*i<= s[0] && s[0] <= -1.0/2.0 +  (1.0/double(nF))*(i+1))     // have to evenly space fibers... 
                        {
//                             std::cout << " i : " << i << std::endl;
//                             printvector(std::cout, s, "s" , "--");
                            // compute Fiber center
                            FieldVector<double,2> Fcenter = { (1.0/(2.0*nF))+((1.0/double(nF))*i)-(1.0/2.0) , 1.0/4.0};
//                             printvector(std::cout, Fcenter, "Fcenter" , "--");
//                             
                            if(std::max( abs(s[0]-Fcenter[0]), abs(s[2]-Fcenter[1])  ) <= rF )
                                output = lambda2;
//                                 return lambda2;                                                         //richtig so? Fiber hat tendenziell größeres mu? 
                            else
                                output = lambda1;
//                                 return lambda1;
                        }
                    }
                }
                else  // lower Layer
                {
                    for(size_t i=0; i<nF ;i++) // running through all the Fibers.. 
                    {
                        if(-1.0/2.0 + (1.0/double(nF))*i<= s[0] && s[0] <= -1.0/2.0 +  (1.0/double(nF))*(i+1))  
                        {
//                             std::cout << " i : " << i << std::endl;
//                             printvector(std::cout, s, "s" , "--");
                            // compute Fiber center
                            FieldVector<double,2> Fcenter = { (1.0/(2.0*nF))+((1.0/double(nF))*i)-(1.0/2.0) , -1.0/4.0 };
//                             printvector(std::cout, Fcenter, "Fcenter" , "--");
                            
                            if(std::max( abs(s[0]-Fcenter[0]), abs(s[2]-Fcenter[1])  ) <= rF )
                                output = lambda2;
//                                 return lambda2;                                                         //richtig so? Fiber hat tendenziell größeres mu? 
                            else
                                output = lambda1;
//                                 return lambda1;
                        }
                    }
                }
                return output;
            };
            return lambdaTerm;
        }
    	else if (imp == "bilayer_lame"){	
		    double lambda1     = parameters.get<double>("mu1",384.615);
		    double lambda2     = parameters.get<double>("mu2",384.615);  
		    
		    auto lambdaTerm = [lambda1, lambda2, phi] (const Domain& z) {
		      if ( isInRotatedPlane(phi, z[dim-2], z[dim-1]) )
		        return lambda1;
		      else
		        return lambda2; };
		    
			return lambdaTerm;
		}

		else if (imp == "helix_poisson"){
		    double E1                = parameters.get<double>("E1", 17e6); //Faser
		    double nu1               = parameters.get<double>("nu1", 0.3);
		    double lambda1 = lameLambda(E1,nu1);

		    double E2                = parameters.get<double>("E2", 17e6); //Polymer, weicher, Querkontraktion groß
		    double nu2               = parameters.get<double>("nu2", 0.5); 
		    double lambda2 = lameLambda(E2, nu2);

		    double r = parameters.get<double>("radius");
			double r2 = parameters.get<double>("radius_helix");

		    auto lambdaTerm = [lambda1, lambda2, r, r2] (const Domain& z) {
		    	std::array<double,2> h = {r*cos(2*M_PI*z[0]), r*sin(2*M_PI*z[0])};
		      	MatrixRT ret(0.0);
		      	if ( sqrt(pow(z[1]-h[0],2)+pow(z[2]-h[1],2)) < r2 )
		        	return lambda1;
		      	else
		        	return lambda2; 
		    };
		    
			return lambdaTerm;
		}

		else if (imp == "chess_poisson"){
		    double E1                = parameters.get<double>("E1", 17e6); //Faser
		    double nu1               = parameters.get<double>("nu1", 0.3);
		    double lambda1 = lameLambda(E1,nu1);

		    double E2                = parameters.get<double>("E2", 17e6); //Polymer, weicher, Querkontraktion groß
		    double nu2               = parameters.get<double>("nu2", 0.5); 
		    double lambda2 = lameLambda(E2, nu2);

		    auto lambdaTerm = [lambda1, lambda2, phi] (const Domain& z) {
		      	if ( (isInRotatedPlane(phi, 	   z[dim-2], z[dim-1]) and isInRotatedPlane(phi + M_PI/2  , z[dim-2], z[dim-1])) or 
		      		 (isInRotatedPlane(phi + M_PI, z[dim-2], z[dim-1]) and isInRotatedPlane(phi + 3*M_PI/2, z[dim-2], z[dim-1])) )
		        	return lambda1;
		      	else
		        	return lambda2; 
		    };
		    
			return lambdaTerm;
		}

// 		else if (imp == "3Dchess_poisson"){
// 		    double E1                = parameters.get<double>("E1", 17e6); //Faser
// 		    double nu1               = parameters.get<double>("nu1", 0.3);
// 		    double lambda1 = lameLambda(E1,nu1);
// 
// 		    double E2                = parameters.get<double>("E2", 17e6); //Polymer, weicher, Querkontraktion groß
// 		    double nu2               = parameters.get<double>("nu2", 0.5); 
// 		    double lambda2 = lameLambda(E2, nu2);
// 
// 		    auto lambdaTerm = [lambda1, lambda2, phi] (const Domain& z) {
// 		    	if (z[0]<0.5)
// 		      		if ( (isInRotatedPlane(phi, 	   z[1], z[2]) and isInRotatedPlane(phi + M_PI/2  , z[1], z[2])) or 
// 		      			 (isInRotatedPlane(phi + M_PI, z[1], z[2]) and isInRotatedPlane(phi + 3*M_PI/2, z[1], z[2])) )
// 		        		return lambda1;
// 		      		else
// 		        		return lambda2; 
// 		        else
// 		        	if ( (isInRotatedPlane(phi,  	   z[1], z[2]) and isInRotatedPlane(phi + M_PI/2  , z[1], z[2])) or 
// 		      			 (isInRotatedPlane(phi + M_PI, z[1], z[2]) and isInRotatedPlane(phi + 3*M_PI/2, z[1], z[2])) )
// 		        		return lambda2;
// 		      		else
// 		        		return lambda1; 
// 		    };
// 		    
// 			return lambdaTerm;
// 		}
// 
// 		else if (imp == "vertical_bilayer_poisson"){ 
//     		double E1                = parameters.get<double>("E1", 17e6); //material1
// 		    double nu1               = parameters.get<double>("nu1", 0.3);
// 		    double lambda1 = lameLambda(E1,nu1);
// 		    double E2                = parameters.get<double>("E2", 17e6); //material2
// 		    double nu2               = parameters.get<double>("nu2", 0.5); 
// 		    double lambda2 = lameLambda(E2, nu2);
// 
// 		    auto lambdaTerm = [lambda1, lambda2, phi] (const Domain& z) {
// 		      if ( isInRotatedPlane(phi+M_PI/2.0, z[dim-2], z[dim-1]) )
// 		        return lambda1;
// 		      else
// 		        return lambda2; };
// 		    
// 			return lambdaTerm;
// 		}
// 
// 		else if (imp == "microstructured_bilayer_poisson"){ 
//     		double E1                = parameters.get<double>("E1", 17e6); //material1
// 		    double nu1               = parameters.get<double>("nu1", 0.3);
// 		    double lambda1 = lameLambda(E1,nu1);
// 		    double E2                = parameters.get<double>("E2", 17e6); //material2
// 		    double nu2               = parameters.get<double>("nu2", 0.5); 
// 		    double lambda2 = lameLambda(E2, nu2);
// 
// 		    auto lambdaTerm = [lambda1, lambda2, phi] (const Domain& z) {
// 		      if ( isInRotatedPlane(phi, z[0]-0.5, z[1]) )
// 		        return lambda1;
// 		      else
// 		        return lambda2; };
// 		    
// 			return lambdaTerm;
// 		}
// 
// 		else { //(imp == "bilayer_poisson")
//     		double E1                = parameters.get<double>("E1", 17e6); //material1
// 		    double nu1               = parameters.get<double>("nu1", 0.3);
// 		    double lambda1 = lameLambda(E1, nu1);
// 		    double E2                = parameters.get<double>("E2", 17e6); //material2
// 		    double nu2               = parameters.get<double>("nu2", 0.5); 
// 		    double lambda2 = lameLambda(E2, nu2);
// 
// 		    auto lambdaTerm = [lambda1, lambda2, phi] (const Domain& z) {
// 		      if ( isInRotatedPlane(phi, z[dim-2], z[dim-1]) )
// 		        return lambda1;
// 		      else
// 		        return lambda2; };
// 		    
// 			return lambdaTerm;
// 		}
	}



};








// TODO  add log here? 


template <int dim>
class PrestrainImp 
{

public:
//     static const int dim = 3;
	static const int dimWorld = 3;
    using CellGridType = YaspGrid< dim, EquidistantOffsetCoordinates< double, dim>>;
//     using GridView = CellGridType::LeafGridView;
    using Domain = typename CellGridType::LeafGridView::template Codim<0>::Geometry::GlobalCoordinate;
//     using Domain = GridView::Codim<0>::Geometry::GlobalCoordinate;
    using MatrixRT = FieldMatrix< double, dimWorld, dimWorld>;
    using ScalarRT = double;
    using Func2Tensor = std::function< MatrixRT(const Domain&) >;
    using FuncScalar = std::function< ScalarRT(const Domain&) >;
    using FuncMatrix = std::function< MatrixRT(const Domain&) >;
    using FuncVector = std::function< VectorRT(const Domain&) >;
protected:
// 	double p1, p2, theta;
// 	double width; //Cell geometry


public:

//     PrestrainImp(double p1, double p2, double theta, double width) : p1(p1), p2(p2), theta(theta), width(width){}

    
    
//     Func2Tensor getPrestrain(std::string imp)
    Func2Tensor getPrestrain(ParameterTree parameters)
    {
        
        std::string imp = parameters.get<std::string>("material_prestrain_imp", "analytical_Example");
        
        double width = parameters.get<double>("width", 1.0);
        double height = parameters.get<double>("height", 1.0);
        
        double theta = parameters.get<double>("theta",1.0/4.0); 
        double p1 = parameters.get<double>("rho1", 1.0);
        double alpha = parameters.get<double>("alpha", 2.0);
        double p2 = alpha*p1;
        
//     	if (imp == "isotropic_bilayer")
//     	{
//             Func2Tensor B = [p1,p2,theta] (const Domain& x)                 //  ISOTROPIC PRESSURE
//             {             
//                 if (abs(x[0]) > (theta/2)  && x[2] > 0)
//                     return MatrixRT{{p1, 0.0 , 0.0}, {0.0, p1, 0.0}, {0.0, 0.0, p1}};
//                 else if (abs(x[0]) < (theta/2)  && x[2] < 0)
//                     return MatrixRT{{p2, 0.0 , 0.0}, {0.0, p2, 0.0}, {0.0, 0.0, p2}};
//                 else
//                     return MatrixRT{{0.0, 0.0 , 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};
// 
//             };
//             std::cout <<" Prestrain Type: isotropic_bilayer " << std::endl;
//             return B;
//         }
        if (imp == "isotropic_bilayer")
    	{
            Func2Tensor B = [p1,p2,theta] (const Domain& x)                 //  ISOTROPIC PRESSURE
            {             
                if (x[2] > 0)
                    return MatrixRT{{p1, 0.0 , 0.0}, {0.0, p1, 0.0}, {0.0, 0.0, p1}};
                else if (x[2] < 0)
                    return MatrixRT{{p2, 0.0 , 0.0}, {0.0, p2, 0.0}, {0.0, 0.0, p2}};
                else
                    return MatrixRT{{0.0, 0.0 , 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};

            };
            std::cout <<" Prestrain Type: isotropic_bilayer " << std::endl;
            return B;
        }
        
        else if (imp == "analytical_Example")
        {
            Func2Tensor B = [p1,theta] (const Domain& x)                // Bilayer with one rectangular Fiber & ISOTROPIC PRESSURE   
            {              
//                 if (abs(x[0]) < (theta/2) && x[2] < 0 )                             //does not make a difference
                if (abs(x[0]) < (theta/2) && x[2] < 0 && x[2] > -(1.0/2.0) )
                    return MatrixRT{{p1, 0.0 , 0.0}, {0.0, p1, 0.0}, {0.0, 0.0, p1}};
                else
                    return MatrixRT{{0.0, 0.0 , 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};
            };
            std::cout <<" Prestrain Type: analytical_Example "<< std::endl;
            return B;
        }
        else if (imp == "parametrized_Laminate")
        {
            Func2Tensor B = [p1,p2,theta] (const Domain& x)           
            {              
                if (abs(x[0]) < (theta/2) && x[2] < 0 )             
                    return MatrixRT{{p2, 0.0 , 0.0}, {0.0, p2, 0.0}, {0.0, 0.0, p2}};
                else if (abs(x[0]) > (theta/2) && x[2] > 0 )    
                    return MatrixRT{{p1, 0.0 , 0.0}, {0.0, p1, 0.0}, {0.0, 0.0, p1}};
                else
                    return MatrixRT{{0.0, 0.0 , 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};
            };
            std::cout <<"Prestrain Type: parametrized_Laminate"<< std::endl;
            return B;
        }
        else if (imp == "circle_fiber"){

            Func2Tensor B = [p1,theta,width] (const Domain& x)                // Bilayer with one rectangular Fiber & ISOTROPIC PRESSURE   
            {  
		      	if (x[0] < 0 && sqrt(pow(x[1],2) + pow(x[2],2) ) < width/4.0  ||  0 < x[0] && sqrt(pow(x[1],2) + pow(x[2],2) ) > width/4.0)
		        	return MatrixRT{{p1, 0.0 , 0.0}, {0.0, p1, 0.0}, {0.0, 0.0, p1}};
		      	else
		        	return MatrixRT{{0.0, 0.0 , 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};
		    };
            std::cout <<" Prestrain Type: circle_fiber"<< std::endl;
            return B;
   		}
        else if (imp == "square_fiber"){
            
            Func2Tensor B = [p1,theta,width] (const Domain& x)                // Bilayer with one rectangular Fiber & ISOTROPIC PRESSURE   
            {  
		      	if (x[0] < 0 && std::max(abs(x[1]), abs(x[2]))  < width/4.0  ||  0 < x[0] && std::max(abs(x[1]), abs(x[2]))  > width/4.0) 
		        	return MatrixRT{{p1, 0.0 , 0.0}, {0.0, p1, 0.0}, {0.0, 0.0, p1}};
		      	else
		        	return MatrixRT{{0.0, 0.0 , 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};
		    };
            std::cout <<" Prestrain Type: square_fiber"<< std::endl;
            return B;
   		}
        else if (imp == "matrix_material_circles")   // Matrix material with prestrained Fiber inclusions
        {
            int nF    = parameters.get<int>("nF", 3);    //number of Fibers in each Layer
            double rF = parameters.get<double>("rF", 0.5*(width/(2.0*nF)) );   //default: half of the max-fiber-radius mrF = (width/(2.0*nF)) 
            
            assert( (2*rF)*nF <= width && (height/4)+rF <= height); //check that fibers are not bigger than Domain
            
            Func2Tensor B = [p1,p2,theta,nF,rF,height,width] (const Domain& x)                
            {              
                //TODO if Domain == 1... if Domain == 2 ...
//                 double domainShift = 1.0/2.0;
                double domainShift = 0.0;
                // shift x to domain [0,1]^3
                FieldVector<double,3> s = {x[0]+domainShift, x[1]+domainShift, x[2]+domainShift};
                
                MatrixRT output;

                // determine if point is in upper/lower Layer
                if ((0.0 <= s[2] && s[2] <= 1.0/2.0)) // upperLayer
                {
                    for(size_t i=0; i<nF ;i++) // running through all the Fibers.. 
                    {
                        if(-1.0/2.0 + (1.0/nF)*i<= s[0] && s[0] <= -1.0/2.0 +  (1.0/nF)*(i+1))  
                        {
                            // compute Fiber center
                            FieldVector<double,2> Fcenter = { (1.0/(2.0*nF))+((1.0/double(nF))*i)-(1.0/2.0) , 1.0/4.0};
//                             
                            if(sqrt(pow(s[0]-Fcenter[0],2)+pow(s[2]-Fcenter[1],2)) <= rF )
                                output = MatrixRT{{p1, 0.0 , 0.0}, {0.0, p1, 0.0}, {0.0, 0.0, p1}};
//                                 return MatrixRT{{p1, 0.0 , 0.0}, {0.0, p1, 0.0}, {0.0, 0.0, p1}};
                            else
                                output = MatrixRT{{0.0, 0.0 , 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};
//                                 return MatrixRT{{0.0, 0.0 , 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};
//                                 return MatrixRT{{p2, 0.0 , 0.0}, {0.0, p2, 0.0}, {0.0, 0.0, p2}};
                        }
                    }
                }
                else  // lower Layer
                {
                    for(size_t i=0; i<nF ;i++) // running through all the Fibers.. 
                    {
                        if(-1.0/2.0 + (1.0/nF)*i<= s[0] && s[0] <= -1.0/2.0 +  (1.0/nF)*(i+1))  
                        {
                            // compute Fiber center
                            FieldVector<double,2> Fcenter = { (1.0/(2.0*nF))+((1.0/double(nF))*i)-(1.0/2.0) , -1.0/4.0};
    
                            if(sqrt(pow(s[0]-Fcenter[0],2)+pow(s[2]-Fcenter[1],2)) <= rF )
                                output = MatrixRT{{p1, 0.0 , 0.0}, {0.0, p1, 0.0}, {0.0, 0.0, p1}};
//                                 return MatrixRT{{p1, 0.0 , 0.0}, {0.0, p1, 0.0}, {0.0, 0.0, p1}};
                            else
                                output = MatrixRT{{0.0, 0.0 , 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};
//                                 return MatrixRT{{0.0, 0.0 , 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};
//                                 return MatrixRT{{p2, 0.0 , 0.0}, {0.0, p2, 0.0}, {0.0, 0.0, p2}};
                        }
                    }

                }
                return output;
            };
            std::cout <<" Prestrain Type: matrix_material"<< std::endl;
            return B;
        }
        
        else if (imp == "matrix_material_squares")   // Matrix material with prestrained Fiber inclusions
        {
            int nF    = parameters.get<int>("nF", 3);    //number of Fibers in each Layer
            double rF = parameters.get<double>("rF", 0.5*(width/(2.0*nF)) );   //default: half of the max-fiber-radius mrF = (width/(2.0*nF)) 
            
            assert( (2*rF)*nF <= width && (height/4)+rF <= height); //check that fibers are not bigger than Domain
            
            Func2Tensor B = [p1,p2,theta,nF,rF,height,width] (const Domain& x)                
            {              
                //TODO if Domain == 1... if Domain == 2 ...
//                 double domainShift = 1.0/2.0;
                double domainShift = 0.0;
                // shift x to domain [0,1]^3
                FieldVector<double,3> s = {x[0]+domainShift, x[1]+domainShift, x[2]+domainShift};
                
                MatrixRT output;

                // determine if point is in upper/lower Layer
                if ((0.0 <= s[2] && s[2] <= 1.0/2.0)) // upperLayer
                {
                    for(size_t i=0; i<nF ;i++) // running through all the Fibers.. 
                    {
                        if(-1.0/2.0 + (1.0/nF)*i<= s[0] && s[0] <= -1.0/2.0 +  (1.0/nF)*(i+1))  
                        {
                            // compute Fiber center
                            FieldVector<double,2> Fcenter = { (1.0/(2.0*nF))+((1.0/double(nF))*i)-(1.0/2.0) , 1.0/4.0};
//                             
                            if(std::max( abs(s[0]-Fcenter[0]), abs(s[2]-Fcenter[1])  ) <= rF )
                                output = MatrixRT{{p1, 0.0 , 0.0}, {0.0, p1, 0.0}, {0.0, 0.0, p1}};
//                                 return MatrixRT{{p1, 0.0 , 0.0}, {0.0, p1, 0.0}, {0.0, 0.0, p1}};
                            else
                                output = MatrixRT{{0.0, 0.0 , 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};
//                                 return MatrixRT{{0.0, 0.0 , 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};
//                                 return MatrixRT{{p2, 0.0 , 0.0}, {0.0, p2, 0.0}, {0.0, 0.0, p2}};
                        }
                    }
                }
                else  // lower Layer
                {
                    for(size_t i=0; i<nF ;i++) // running through all the Fibers.. 
                    {
                        if(-1.0/2.0 + (1.0/nF)*i<= s[0] && s[0] <= -1.0/2.0 +  (1.0/nF)*(i+1))  
                        {
                            // compute Fiber center
                            FieldVector<double,2> Fcenter = { (1.0/(2.0*nF))+((1.0/double(nF))*i)-(1.0/2.0) , -1.0/4.0};
    
                            if(std::max( abs(s[0]-Fcenter[0]), abs(s[2]-Fcenter[1])  ) <= rF )
                                output = MatrixRT{{p1, 0.0 , 0.0}, {0.0, p1, 0.0}, {0.0, 0.0, p1}};
//                                 return MatrixRT{{p1, 0.0 , 0.0}, {0.0, p1, 0.0}, {0.0, 0.0, p1}};
                            else
                                output = MatrixRT{{0.0, 0.0 , 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};
//                                 return MatrixRT{{0.0, 0.0 , 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};
//                                 return MatrixRT{{p2, 0.0 , 0.0}, {0.0, p2, 0.0}, {0.0, 0.0, p2}};
                        }
                    }

                }
                return output;
            };
            std::cout <<" Prestrain Type: matrix_material"<< std::endl;
            return B;
        }
        else   
        {
            // TODO other geometries etc... 
            
        }
        // TODO ANISOTROPIC PRESSURE
        

    }
    
    
    
    FuncScalar getNormPrestrain(const FuncMatrix& B)
    {
	    FuncScalar normB = [&](const Domain& z) {
	      	return norm(B(z)); 
	    };
	    return normB;  
    }


};







#endif
